'----------------------------------------------------------------------------------------------------
'Developed by Charles Dietz, 11-03-2008
'This include file defines a class to create an unsorted listbox control consisting of columns.
'Headers for each column are defined and can be clicked to toggle the sorting of that column.
'
'The handle of the control's parent is denoted by hDlg. 
'The control itself is identified by listID, while the headers are identified by listID + 1, 
'listID + 2, listID + 3, ...  Therefore, the listID's of controls must be spaced enough to allow
'for the control's columns.  Controls spaced at multiples of 10 would allow for up to 9 columns.
'Controls are referred to as columnLists and columnLists controls are handled here as objects.
'multiple columnList controls can be created, as shown here for two controls
'ColumnList controls can be created in terms of either pixels or units. 
'recNo() keeps track of records as columns are sorted for internal display purposes
'extra() holds extra data for each item used by calling program for whatever purpose.
'----------------------------------------------------------------------------------------------------

'#INCLUDE ONCE "WIN32API.INC"
'#INCLUDE ONCE "COMDLG32.INC"

CLASS cColumnList
   INSTANCE hDlg, listID AS LONG       'handle of column listboxes parent
   INSTANCE numCols AS LONG            'number of columns in listbox
   INSTANCE numItems AS LONG           'number of items in listbox
   INSTANCE xList, yList AS LONG       'listbox position (upper left corner)
   INSTANCE hiLite AS LONG             'high lighted item in listbox
   INSTANCE wi() AS LONG               'column header widths (in percent, neg for right alligned)
   INSTANCE colType() AS LONG          '1,2,3 for str, val, date column type (neg for right-adjust) 
   INSTANCE items() AS STRING          'row strings with TAB separated columns
   INSTANCE recNo() AS LONG            'record number as originally added (sorting purposes)
   INSTANCE extra() AS LONG            'extra LONG value (for purposes of calling program)
   INSTANCE nPixels AS LONG

   INTERFACE iColumnList '---------------------------------------------------------------------
      INHERIT IUNKNOWN

      PROPERTY GET getHandle() AS LONG 
         PROPERTY = hDlg 
      END PROPERTY

      PROPERTY GET getlistID() AS LONG
         PROPERTY = listID
      END PROPERTY
      
      PROPERTY GET getNumCols() AS LONG
         PROPERTY = numCols 
      END PROPERTY

      PROPERTY GET getNumItems() AS LONG 
         PROPERTY = numItems 
      END PROPERTY
      
      PROPERTY SET setNumItems(BYVAL n AS LONG)
         numItems = n
      END PROPERTY

      PROPERTY GET hiLite AS LONG
         IF numItems THEN PROPERTY = hiLite 
      END PROPERTY
      
      METHOD setHiLite(BYVAL n AS LONG, OPT BYVAL noDraw AS LONG, BYVAL top AS LONG)
         'n = item number to be hiLited
         'nodraw = 1 if redraws are to be supressed until done
         'top = item number at top of listbox window
         '--------------------------------------------------------------------
         IF noDraw THEN CONTROL SEND hDlg, listID, %WM_SETREDRAW, 0, 0
         LISTBOX SELECT hDlg, listID, n
         CONTROL SEND hDlg, listID, %LB_SETTOPINDEX, top - 1, 0
         CONTROL SEND hDlg, listID, %WM_SETREDRAW, 1, 0
         hiLite = n
      END METHOD

      METHOD showList()
         LOCAL i, k AS LONG, s AS STRING
         CONTROL SEND hDlg, listID, %WM_SETREDRAW, 0, 0
         LISTBOX RESET hDlg, listID
         FOR i = 1 TO numItems
            s = items(recNo(i))
            LISTBOX ADD hDlg, listID, s
         NEXT i
         CONTROL SEND hDlg, listID, %WM_SETREDRAW, 1, 0
      END METHOD

      METHOD getRecordNo(BYVAL i AS LONG) AS LONG 
         IF UBOUND(recNo()) >= i THEN METHOD = recNo(i)
      END METHOD                                                        
      
      METHOD getExtra(BYVAL i AS LONG) AS LONG
         IF UBOUND(extra()) >= i THEN METHOD = extra(i)
      END METHOD 
      
      METHOD getListboxPos(x AS LONG, y AS LONG)
         'return listbox upper left corner position
         x = xList: y = yList
      END METHOD
      
      METHOD createCtrl(hParent AS LONG, ID AS LONG, _
                        a AS LONG, b AS LONG, w AS LONG, h AS LONG, _
                        pwc AS DWORD, ptc AS DWORD, ncols AS LONG)
         'hParent, id      = handle of containing dialog, control id
         'a, b, w, h       = location and size of control including listbox headers (pixels/units from parent)
         'pwc              = pointer to wc(), width of columns in percent (neg value for right adjusted)
         'ptc              = pointer to tc(), title of columns (leading chr$(1) for value, chr$(2) for date)
         'ncols            = number of columns
         '-----------------------------------------------------------------------------------------------------     
         LOCAL nStyle, i, j, k, hdc, ws, ht, rx, ry, x, y, xx, yy AS LONG
         LOCAL cSize AS SIZEL, r AS RECT, txt AS ASCIIZ*16, s AS STRING, f AS DOUBLE
         LOCAL wwi() AS LONG
         
         numCols = ncols
         DIM wc(1 TO numCols) AS LONG AT pwc
         DIM tc(1 TO numCols) AS STRING AT ptc
         DIM wi(1 TO numCols) AS INSTANCE LONG
         DIM wwi(1 TO numCols) AS LONG
         DIM tabstops(numCols - 1) AS LONG
         DIM colType(1 TO numCols) AS INSTANCE LONG
         
         FOR i = 1 TO numCols: wi(i) = wc(i): NEXT i
         'determine if pixels or dialog units ---------------------
         hDlg = hParent: listID = ID 
         DIALOG GET SIZE hDlg TO rx, ry  
         GetWindowRect  hDlg, r
         IF rx = (r.nRight - r.nLeft) THEN nPixels = 1
         
         'determine header height and scroll width ----------------
         txt = "0": hdc = GetDC(hDlg)
         GetTextExtentPoint32 hdc, BYVAL VARPTR(txt), LEN(txt), cSize
         ht = 1.2 * cSize.cy
         ws = GetSystemMetrics(%SM_CXVSCROLL)
         IF nPixels = 0 THEN DIALOG PIXELS hDlg, ws, ht TO UNITS ws, ht
         ReleaseDC hDlg, hdc
         FOR i = 1 TO numCols: f = wi(i) * (w - ws - 8) / 100: wi(i) = f: NEXT i
         
         'define column types -------------------------------------
         FOR i = 1 TO numCols
            colType(i) = ASC(LEFT$(tc(i), 1)) + 1
            IF colType(i) > 3 THEN colType(i) = 1
            colType(i) = SGN(wi(i)) * coltype(i): wwi(i) = ABS(wi(i))
         NEXT i

         'fit last column width -----------------------------------
         j = 0: FOR i = 1 TO numCols - 1: j = j + wwi(i): NEXT i
         wwi(numCols) = w - j - ws - 8
         
         'create the listbox component of the control -------------
         x = a: y = b: xx = w: yy = h: xList = x: yList = y + ht
         nStyle = %WS_VSCROLL OR %LBS_DISABLENOSCROLL OR %LBS_USETABSTOPS OR %WS_TABSTOP 'not sorted 
         CONTROL ADD LISTBOX, hDlg, listID, , xList, yList, xx-8, yy-ht-10, nStyle, %WS_EX_CLIENTEDGE, CALL ListCB
         CONTROL SET USER hDlg, listID, 1, OBJPTR(ME)
         
         'define tabstops for listbox -----------------------------
         tabstops(0) = 2
         FOR i = 1 TO numCols - 1: tabstops(i) = tabstops(i - 1) + wwi(i): NEXT i
         tabstops(numCols - 1) = tabstops(numCols - 1) - 2
         FOR i = 0 TO numCols - 1
            IF colType(i + 1) < 0 THEN 'right alligned tabstop
               tabstops(i) = SGN(colType(i + 1)) * (tabstops(i) + wwi(i + 1) - 6)
            END IF
         NEXT i
         IF nPixels THEN 'convert pixels to dialog units for tabstops
            DIALOG PIXELS hDlg, 1000, 0 TO UNITS rx, ry
            FOR i = 0 TO numCols - 1: tabstops(i) = rx/1000 * tabstops(i): NEXT i
         END IF
         CONTROL SEND hDlg, listID, %LB_SETTABSTOPS, numCols, VARPTR(tabstops(0))

         'create header components of the control -----------------
         j = x
         FOR i = 1 TO numCols 
            IF i > 1 THEN j = j + wi(i - 1)
            s = tc(i): IF ABS(colType(i)) > 1 THEN s = MID$(s, 2)
            IF colType(i) < 0 THEN 
               s = s + "  ": nStyle = %SS_RIGHT + %SS_NOTIFY 
            ELSE                                                    
               s = "  " + s: nStyle = %SS_LEFT + %SS_NOTIFY
            END IF
            CONTROL ADD LABEL, hDlg, listID+i, s, j, y, wwi(i), ht, nStyle, %WS_EX_STATICEDGE, CALL ListCB
            CONTROL SET USER hDlg, listID+i, 1, OBJPTR(ME)
         NEXT i
      END METHOD
      
      METHOD addItem(itemStr AS STRING, OPT BYVAL nExtra AS LONG)
         'add itemStr to items()                                                      
         'truncates any column text that is too long
         '----------------------------------------------------------------
         LOCAL i, x, y, px, py, hList AS LONG, s AS STRING
         DIM sCols(1 TO numCols) AS STRING
         PARSE itemStr + $tab, sCols(), $TAB
         CONTROL HANDLE hDlg, listID TO hList
         UnitsToPixels(hDlg, x, y, px, py)                                                                
         FOR i = 1 TO numCols
            x = .95 * ABS(wi(i))                          
            IF nPixels = 0 THEN 
               'UnitsToPixels(hDlg, x, y, px, py)
               DIALOG UNITS hMain, x, y TO PIXELS px, py
               IF %PB_REVISION = &H1004  then 'compiler PBWIN10.04                  
                   SWAP px, py
               END IF                  
            end if
            ME.truncateStr(hList, sCols(i), px)                                                            
            s = s + sCols(i) + $TAB                          
         NEXT i
         itemStr = RTRIM$(s, $TAB)
         INCR numItems
         REDIM PRESERVE items(1 TO numItems)           
         REDIM PRESERVE recNo(1 TO numItems)
         REDIM PRESERVE extra(1 TO numItems)
         items(numItems) = itemStr  
         recNo(numItems) = numItems
         extra(numItems) = nExtra
      END METHOD

      METHOD truncateStr(hCtrl AS LONG, s AS STRING, allowLen AS LONG)
         'hCtrl = handle of control, allowLen = allowable length in pixels
         LOCAL n, nFit, hdc, hFont, hFontOld AS LONG
         LOCAL sz AS ASCIIZ*128, cSize AS SIZEL
         hdc = GetDC(hCtrl): sz = s: n = LEN(s)                    
         hFont = SendMessage(hCtrl, %WM_GETFONT, 0, 0)
         hFontold = SelectObject(hDC, hFont)
         GetTextExtentExPoint hdc, sz, n, allowLen, nFit, BYVAL 0, cSize
         s = LEFT$(s, nFit)
         SelectObject(hDC, hFontOld)
         ReleaseDC hCtrl, hdc
      END METHOD

      METHOD convertDate(s AS STRING) AS LONG
         'convert "10/13/1934" to 19341013 
         LOCAL mo AS WORD, day AS WORD, yr AS WORD
         mo = VAL(LEFT$(s, 2))
         day = VAL(MID$(s, 4, 2))
         yr = VAL(RIGHT$(s, 4)) - 1980
         ROTATE RIGHT yr, 7
         ROTATE LEFT mo, 5
         METHOD = yr + mo + day
      END METHOD
      
      METHOD columnSort(nCol AS LONG)
         'this routine toggles the column sort when the column header is clicked.
         'colType() holds the type of column: string=1, value=2, date=3
         'tagged QUAD array temp() contains both recNo() and extra() 
         'the tagged array recNo() is used to display the sorted columns
         '------------------------------------------------------------------------
         LOCAL i, n, nSort AS LONG, s AS STRING
         DIM sArray(1 TO numItems) AS STRING
         DIM nArray(1 TO numItems) AS LONG
         DIM temp(1 TO numItems) AS QUAD
         IF numItems = 0 THEN EXIT METHOD
         hiLite = ME.getRecordNo(hiLite)
         n = ABS(colType(nCol)) 'neg colType signals right adjusted
         FOR i = 1 TO numItems 'fill the appropriate array with the specified column
            s = PARSE$(items(recNo(i)), $TAB, nCol)
            IF n = 2 THEN 'value, use numeric array
               nArray(i) = VAL(s)
            ELSEIF n = 3 THEN 'date, use numeric array
               nArray(i) = ME.convertDate(s)
            ELSE 'use string array
               sArray(i) = s
            END IF   
         NEXT i
         nSort = -1
         FOR i = 1 TO numItems
            temp(i) = MAK(QUAD, recNo(i), extra(i))
         NEXT i   
         IF sArray(numItems) > sArray(1) OR nArray(numItems) > nArray(1) THEN nSort = 1
         IF nSort < 0 THEN 'sort as ascending
            IF n = 1 THEN 'use string array
               ARRAY SORT sArray(), COLLATE UCASE, TAGARRAY temp(), ASCEND
            ELSE 'use numeric array
               ARRAY SORT nArray(), TAGARRAY temp(), ASCEND
            END IF 
         ELSE 'sort as descending
            IF n = 1 THEN 'use string array
               ARRAY SORT sArray(), COLLATE UCASE, TAGARRAY temp(), DESCEND
            ELSE 'use numeric array
               ARRAY SORT nArray(), TAGARRAY temp(), DESCEND
            END IF
         END IF 
         FOR i = 1 TO numItems 
            recNo(i) = LO(LONG, temp(i))
            extra(i) = HI(LONG, temp(i))
         NEXT i
         ME.showList()
         'n = hiLite  'ME.getHiLite
         'ME.setHiLite = n
         ME.setHiLite(hiLite)
      END METHOD
   END INTERFACE   
END CLASS

CALLBACK FUNCTION listCB()
   'callback for objects listbox
   LOCAL m, n, i, j, k, numItems, idLB AS LONG
   LOCAL iObj AS iColumnList, meAddr AS DWORD
   
   IF CB.CTLMSG = %STN_CLICKED OR CB.CTLMSG = %LBN_SELCHANGE THEN
      CONTROL GET USER CB.HNDL, CB.CTL, 1 TO meAddr
      POKE DWORD, VARPTR(iObj), meAddr
      IF ISOBJECT(iObj) THEN iObj.AddRef ELSE EXIT FUNCTION
      idLB = iObj.getlistID
      IF CB.CTL = idLB AND CB.CTLMSG = %LBN_SELCHANGE THEN 'record hiLite
         LISTBOX GET SELECT CB.HNDL, CB.CTL TO n
         iObj.setHiLite(MAX(n, 1))
      ELSEIF CB.CTL <> idLB AND CB.CTLMSG = %STN_CLICKED THEN 'sort column
         n = iObj.hiLite  'getHiLite 
         m = iObj.getRecordNo(n)
         iObj.columnSort(CB.CTL - idLB)
         numItems = iObj.getNumItems
         FOR i = 1 TO numItems
            IF iObj.getRecordNo(i) = m THEN j = i: EXIT FOR
         NEXT i
         iObj.setHiLite(j)
      END IF
   END IF
END FUNCTION

